/*******************************************************************************


	File        :   cshade.c

	Description :   CyberShades glasses driver implementation

	Version     :   1.01

	Copyright   :   1995 Chinon America, Inc. All rights reserved.

					The code contained therein is copyrighted
by
					Chinon America, which reserves its rights
for
					distribution and use of this product.
You
					should first consult with Chinon America
before
					modifying it, distributing it or copying
some
					or all of its content.

*******************************************************************************/

#include <dos.h>
#include <stdlib.h>
#include <conio.h>

#include "cshade.h"

#define outportb outp
#define inportb inp
#define setvect _dos_setvect
#define getvect _dos_getvect

#define _CSHADE_IMP_VERSION       (0x101)

#if (_CSHADE_IMP_VERSION != _CSHADE_VERSION)
 #error Header file 'cshade.h' is wrong version.
#endif

/*******************************************************************************
**
**  Port i/o implementation
**  =======================
**
**
**  Get port address (COM ports : address valid for glasses i/o ONLY)
**
*/

#define DosPortAddress ((unsigned short *) 0x400)

	unsigned int PortAddress (StandardPort p) {

		unsigned int result = 0;

		if (IsComPort(p) || IsLptPort(p)) {

			result = DosPortAddress [p];

			if (result && IsComPort(p)) result += 4;
		}

		return result;
	}

/***************************************
**
**  Read from port
**
*/
	char PortRead (StandardPort p) {
		return inportb (PortAddress(p));
	}

/***************************************
**
**  Write to a port
*/
	void PortWrite (StandardPort p, char byte) {
		outportb (PortAddress(p), byte);
	}


/*******************************************************************************
**
**  Glasses control
**  ===============
**
**
**  Initialize glasses data given a port and a CyberShades device
**
*/
	int InitCyberShadesGlasses (   GlassesData* data,
							_CyberShadesDevice
device,
							StandardPort port)
	{

	/*  make sure port exists
	*/
		if (!PortExists(port)) return 0;

	/*  set device data
	*/
		switch (device) {

		case PCP:       /*  port must be parallel
*/
			if (!IsLptPort(port)) return 0;
			InitCustomGlasses (data, port, PCP_LEFTMASK,
PCP_RIGHTMASK, 0);
			break;

		case PCS:       /*  port must be serial
*/
			if (!IsComPort(port)) return 0;
			InitCustomGlasses (data, port, PCS_LEFTMASK,
PCS_RIGHTMASK, 0);
			break;

		case PC3D:      /*  port must be serial
*/
			if (!IsComPort(port)) return 0;
			InitCustomGlasses (data, port, PC3D_LEFTMASK,
PC3D_RIGHTMASK, 0);
			break;

		case SS1:       /*  COM or LPT connection
*/
			if (IsComPort(port)) {
				InitCustomGlasses (data, port, SS1_COM_LEFTMASK,
SS1_COM_RIGHTMASK, 0);
			} else {
				InitCustomGlasses (data, port, SS1_LPT_LEFTMASK,
SS1_LPT_RIGHTMASK, 0);
			}
			break;


		case PCVR:      /*  can be COM or LPT connection
*/
			if (IsComPort(port)) {
				InitCustomGlasses (data, port, PCVR_COM_LEFTMASK,
PCVR_COM_RIGHTMASK, 0);
			} else {
				InitCustomGlasses (data, port, PCVR_LPT_LEFTMASK,
PCVR_LPT_RIGHTMASK, 0);
			}
			break;

		default:        /*  error, unknown device
*/
			return 0;
		}

		return 1;
	}


/***************************************
**
**  Initialize glasses with custom masks
**
*/
	int InitCustomGlasses (GlassesData* data, StandardPort port,
							char leftMask, char
rightMask, char offMask)
	{
	/*  make sure port exists
	*/
		if (!PortExists(port)) return 0;

	/*  set device data
	*/
		data->Port = port;
		data->LeftMask = leftMask;
		data->RightMask = rightMask;
		data->OffMask = offMask;
		data->Reversed = 0;

	/*  Turn glases off
	*/
		TellGlasses (data, Off);

	/*  return OK
	*/
		return 1;
	}

/***************************************
**
**  Send commands to the glasses
**  ============================
*/
	void TellGlasses (GlassesData* data, GlassesCommand command)
	{

		char temp;

	/*  check port is valid
	*/
		if (!PortExists(data->Port)) return;

	/*  Switch reverses left and right masks
	*/
		if (command == Switch ) {
			data->Reversed = !data->Reversed;
		}

	/*  read port and mask so as to not overwrite
	**  unused bits that may be otherwise used (PowerGlove, etc...)
	*/
		temp = PortRead (data->Port);
		temp &= ~(data->LeftMask | data->RightMask | data->OffMask);

	/*  act on command and update data->State
	*/
		if (command == Flip) {
			if (data->State == Left) command = (!data->Reversed) ?
Right : Left;
			if (data->State == Right) command = (!data->Reversed) ?
Left : Right;
		}

		switch (command) {

		case Off:
			PortWrite (data->Port, temp | data->OffMask);
			data->State = Off;
			break;

		case Left:
			PortWrite (data->Port, temp | data->LeftMask);
			data->State = (!data->Reversed) ? Left : Right;
			break;

		case Right:
			PortWrite (data->Port, temp | data->RightMask);
			data->State = (!data->Reversed) ? Right : Left;
			break;
		}
	}

/*******************************************************************************
**
**  Timer interrupt
**  ===============
**
**  Constants
**  =========
*/

#define TIMER_INT       8               /*  Timer interrupt vector          */

#define TIMER_MODE      0x34            /*  Timer control codes             */
#define TIMER_CONTROL   0x43
#define TIMER_0         0x40
#define TIMER_LATCH     0x00

#define PIC_CMD         0x20
#define NONSPEC_EOI     0x20

#define RESYNC_COUNT        50          /* frames between resync            */
#define INTERRUPT_ADVANCE   1000        /* clock advance in clock ticks     */

/****************************************
**
**  Data types
**  ==========
*/
	typedef void interrupt (INT_VECTOR)(void);
	typedef void (USERTIMERPROC)(void);

/***************************************
**
**  Static data
**  ===========
*/
	static unsigned int FrameCounter;   /* used to generate re-sync
*/
	static unsigned int FrameDelay;     /*  delay between frames
*/
	static INT_VECTOR far * DosTick;         /* original timer handler
*/
	static USERTIMERPROC* UserProc = 0; /* user- defined function
*/

/***************************************
**
**  Waits for the start of the next vertical retrace
**
*/
	#define V_RETRACE           ((inportb (0x03DA) & 0x08) != 0)

	static void WaitForRetrace (void) {
		while (V_RETRACE);
		while (!V_RETRACE);
	}

/***************************************
**
**  Read system timer
**  =================
*/
	static unsigned int GetTimer (void) {
		outportb (TIMER_CONTROL, TIMER_LATCH);
		return (inportb (TIMER_0) + (inportb (TIMER_0) << 8));
	}

/***************************************
**
**  Set system timer interrupt delay
**  ================================
*/
	static void SetTimer (unsigned int t) {
		outportb (TIMER_CONTROL, TIMER_MODE);
		outportb (TIMER_0, t & 0xFF);
		outportb (TIMER_0, t >> 8);
	}

/***************************************
**
**  Clock the VGA vertical sync
**  ===========================
**
*/
	static int TimeVerticalScan (void) {

		unsigned int start, stop;

		disable();
	/*  sync with VGA
	*/
		WaitForRetrace();

	/*  avoid timer interrupt while clocking
	*/
		SetTimer (0xFFFF);

		WaitForRetrace();
		start = GetTimer();
		WaitForRetrace();
		WaitForRetrace();
		stop = GetTimer();

		enable();

		return (start - stop) >> 1;
	}

/***************************************
**
**  The timer interrupt routine
**  ===========================
*/
	static void interrupt TimerHandler (void) {

	/*  Data
	*/
		static unsigned int last_tick;
		unsigned int tick_count = FrameDelay;

		disable();

	/*  call user procedure
	*/
		if (UserProc) UserProc();

	/*  we are a little early, so change time after re-sync
	*/
		if (FrameCounter == RESYNC_COUNT) {
			SetTimer (FrameDelay);
			tick_count -= INTERRUPT_ADVANCE;
		}

	/*  re-sync
	*/
		if (--FrameCounter == 0) {

			SetTimer (0xFFFF);
			WaitForRetrace();
			SetTimer (FrameDelay - INTERRUPT_ADVANCE);
			FrameCounter = RESYNC_COUNT;
		}

	/*  simulate DOS tick
	*/
		tick_count += last_tick;
		if (tick_count < last_tick) DosTick();
		last_tick = tick_count;

	/*  clear interrutp controller
	*/
		outportb(PIC_CMD, NONSPEC_EOI);
		enable();
	}

/***************************************
**
**  Uninstall the timer interrupt
**  =============================
*/
	static void StopTimer (void) {

	/*  check if installed
	*/
		if (!UserProc) return;

	/*  restore vector
	*/
		disable();
		SetTimer (0x0000);
		setvect (TIMER_INT, DosTick);
		enable();

	/*  reset traps
	*/

		UserProc = 0;
	}

/***************************************
**
**  Install the timer interrupt
**  ===========================
*/
	static int StartTimer (USERTIMERPROC* proc) {

	/*  check if already installed
	*/
		if (UserProc || !proc) return 0;


	/*  set traps in case user forgets to uninstall or program error
	*/
		atexit (StopTimer);

	/*  set handler data
	*/
		FrameDelay = TimeVerticalScan();
		UserProc = proc;
		FrameCounter = RESYNC_COUNT;

		disable();
		DosTick = getvect (TIMER_INT);
		WaitForRetrace();
		setvect (TIMER_INT, TimerHandler);
		SetTimer (FrameDelay - INTERRUPT_ADVANCE);
		enable();

	/*  return OK
	*/
		return 1;
	}

/*******************************************************************************
**
**  Glasses installation
**  ====================
**
**  Static data
**  ===========
*/
	static VGAFlipProc* VGAFlip = 0;/*  user-defined page flipping
procedure*/
	static GlassesData* Glasses;    /*  user-supplied glasses data
*/
	static int nGlasses;            /*  number of glasses to drive
*/


/***************************************
**
**  Glasses procedure, flips sides
**  ==============================
*/
	static void GlassesProc (void) {

		int i;

	/*  send Flip command to all glasses being driven
	*/
		for (i = 0; i < nGlasses; i++)
			TellGlasses (&Glasses[i], Flip);

	/*   tell user which page to show
	*/
		VGAFlip (Glasses[0].State);
	}


/***************************************
**
**  Start the glasses
**  =================
*/
	int StartGlasses (GlassesData* data, VGAFlipProc* proc,
						GlassesCommand firstSide)
	{
		return StartMultipleGlasses (data, 1, proc, firstSide);
	}

/***************************************
**
**  Start multiple glasses
**  ======================
*/
	int StartMultipleGlasses (GlassesData* data, int n, VGAFlipProc proc,
								GlassesCommand
firstSide)
	{

		int i;

	/*  check if already installed
	*/
		if (VGAFlip || !proc) return 0;

	/*  set data
	*/
		VGAFlip = proc;
		Glasses = data;
		nGlasses = n;

	/*  start the interrupt handler
	*/
		if (!StartTimer (GlassesProc)) return 0;

	/*  Set first side
	*/
		for (i = 0; i < n; i++)
			TellGlasses (&data[i], firstSide);

	/*  return OK
	*/
		return 1;
	}

/***************************************
**
**  Stop driving the glasses
**  ========================
*/
	void StopGlasses (void) {

		int i;

	/*  check if installed
	*/
		if (!VGAFlip) return;

	/*  stop the timer handler
	*/
		StopTimer();

	/*  turn off the glasses
	*/
		for (i = 0; i < nGlasses; i++)
			TellGlasses (&Glasses[i], Off);

	/*  clear static data
	*/
		VGAFlip = 0;
	}


/*******************************************************************************
**
**	VGA frequency increase
**	======================
*/
	#define MISC_OUT_WRITE  0x3C2
	#define MISC_OUT_READ   0x3CC

	int IncreaseVGAFreq () {

		char disp_freq;
		int old_delay = TimeVerticalScan();

		/* increase clock frequency */
		outportb (MISC_OUT_WRITE, (disp_freq = inportb (MISC_OUT_READ)) |
4);

		if (TimeVerticalScan () > old_delay * 0.9)
		{
			outportb (MISC_OUT_WRITE, disp_freq);
			return 0;
		}
		return 1;
	}

/* eof ************************************************************************/



