// joystick.c
// Copyright 1995 John Schultz
// All Rights Reserved
// Created 9/19/93
// Modified 12/16/95

#include <conio.h>

#include "joystick.h"
#include "vtimer.h"
#include "lcdcall.h"

#define PIC_MASK1 0x21
#define PIC_MASK2 0xA1

#define JOY_PORT  0x201

// Bit fields for joyButtons:
#define JOY_A_B1    (1<<0)
#define JOY_A_B2    (1<<1)
#define JOY_B_B1    (1<<2)
#define JOY_B_B2    (1<<3)
#define BUTTON_MASK (JOY_A_B1|JOY_A_B2|JOY_B_B1|JOY_B_B2)

#define JOY_MAX 2048   // Maximum possible read value (user adjustable).

long joyMax = JOY_MAX; // joyMax can be used to check whether to disable
                       // a non-existant joystick axis by modifying the
                       // joyflags variable (joystick axis read mask).
                       // Also, joyMax can be set to the maximum pot
                       // value read during calibration to eliminate
                       // spurious over-reads.

// Set this flag to 1 in all interrupt code to let readJoystick() know
// to throw out the current reading.
volatile long interrupted;

static long joyflags=0,intMask[2]={0,0};

void initJoystick(long flags,long joyMaxRead,long intMaskSet1,long intMaskSet2) {
  joyflags = flags;
  joyMax   = joyMaxRead;
  intMask[0] = ~intMaskSet1;
  intMask[1] = ~intMaskSet2;
} // initJoystick

#define MAX_JOY_TRIES 4

// Returns true if joystick read successfully without being interrupted.

long readJoystick(joyInfo * ji) {
long oldIntMask[2],i;
long oJoyX1 = joyX1,oJoyY1 = joyY1,oJoyX2 = joyX2,oJoyY2 = joyY2;
long oldDelayCount;

  for (i=0; i < MAX_JOY_TRIES; i++) {

    oldIntMask[0] = inp(PIC_MASK1);
    oldIntMask[1] = inp(PIC_MASK2);

    outp(PIC_MASK1,intMask[0]); // Block (mask) all but serial interrupts & timer.
    outp(PIC_MASK2,intMask[1]); // Block (mask) all but serial interrupts & timer.

    interrupted = 0; // All interrupt routines must set this var to 1.
    oldDelayCount = (volatile)LCDFlagsP->delayCount;

    readJoy(joyflags); // Call optimized assembly joystick reader.

    outp(PIC_MASK2,oldIntMask[1]); // Restore interrupt masks
    outp(PIC_MASK1,oldIntMask[0]); // Restore interrupt masks

    if (!interrupted && ((volatile)LCDFlagsP->delayCount == oldDelayCount)) break;

  } // for

  ji->js[0].button[0] = (joyButtons & JOY_A_B1) != 0;
  ji->js[0].button[1] = (joyButtons & JOY_A_B2) != 0;

  ji->js[1].button[0] = (joyButtons & JOY_B_B1) != 0;
  ji->js[1].button[1] = (joyButtons & JOY_B_B2) != 0;

  if (i == MAX_JOY_TRIES) {
    ji->js[0].x = oJoyX1;
    ji->js[0].y = oJoyY1;
    ji->js[1].x = oJoyX2;
    ji->js[1].y = oJoyY2;
    return 0;
  } else {
    ji->js[0].x = joyX1;
    ji->js[0].y = joyY1;
    ji->js[1].x = joyX2;
    ji->js[1].y = joyY2;
    if ((joyflags & JOY_A_X) && joyX1 >= joyMax) return 0;
    if ((joyflags & JOY_A_Y) && joyY1 >= joyMax) return 0;
    if ((joyflags & JOY_B_X) && joyX2 >= joyMax) return 0;
    if ((joyflags & JOY_B_Y) && joyY2 >= joyMax) return 0;
    return 1;
  } // if

} // readJoystick

#define TIMER_READ_TRIES 2

long readJoystickTimer(joyInfo * ji) {
long oldIntMask[2],i;
long oJoyX1 = joyX1,oJoyY1 = joyY1,oJoyX2 = joyX2,oJoyY2 = joyY2;
long t1,jt[4];
long oldDelayCount,newDelayCount;
long oldBWaitCount,newBWaitCount;
long tries;

  joyX1 = joyY1 = joyX2 = joyY2 = 0;

  oldIntMask[0] = inp(PIC_MASK1);
  oldIntMask[1] = inp(PIC_MASK2);

// Read joystick axies
  for (tries = 0; tries < TIMER_READ_TRIES; tries++) {

    vdata.readingJoystick = 1; // Don't allow calls to dos tick or sound code

    outp(PIC_MASK1,intMask[0]); // Block (mask) all but serial interrupts & timer.
    outp(PIC_MASK2,intMask[1]); // Block (mask) all but serial interrupts & timer.

    jt[0] = jt[1] = jt[2] = jt[3] = 0;

    oldDelayCount = (volatile)LCDFlagsP->delayCount;
    oldBWaitCount = (volatile)vdata.busyWaitCount;

    outp(JOY_PORT,0); // Init joystick for reading

    t1 = getMicroSecs();

    for (i=0; i < joyMax; i++) {
      long bits = inp(JOY_PORT) & joyflags;
      if (!jt[0] && !(bits & (1<<0))) jt[0] = getMicroSecs();
      if (!jt[1] && !(bits & (1<<1))) jt[1] = getMicroSecs();
      if (!jt[2] && !(bits & (1<<2))) jt[2] = getMicroSecs();
      if (!jt[3] && !(bits & (1<<3))) jt[3] = getMicroSecs();
      if (!bits) break;
    } // while

    newDelayCount = (volatile)LCDFlagsP->delayCount;
    newBWaitCount = (volatile)vdata.busyWaitCount;

    outp(PIC_MASK2,oldIntMask[1]); // Restore interrupt masks
    outp(PIC_MASK1,oldIntMask[0]); // Restore interrupt masks

    vdata.readingJoystick = 0; // Allow calls to dos tick or sound code

    if (oldDelayCount == newDelayCount && oldBWaitCount == newBWaitCount) break;

  } // for

// Read buttons
  joyButtons = inp(JOY_PORT);
  joyButtons = (~joyButtons)>>4;

  if (joyflags & JOY_A_X) {
    joyX1 = jt[0] - t1;
    if (joyX1 < 0) joyX1 = oJoyX1;
  } // if
  if (joyflags & JOY_A_Y) {
    joyY1 = jt[1] - t1;
    if (joyY1 < 0) joyY1 = oJoyY1;
  } // if
  if (joyflags & JOY_B_X) {
    joyX2 = jt[2] - t1;
    if (joyX2 < 0) joyX2 = oJoyX2;
  } // if
  if (joyflags & JOY_B_Y) {
    joyY2 = jt[3] - t1;
    if (joyY2 < 0) joyY2 = oJoyY2;
  } // if

  ji->js[0].button[0] = (joyButtons & JOY_A_B1) != 0;
  ji->js[0].button[1] = (joyButtons & JOY_A_B2) != 0;

  ji->js[1].button[0] = (joyButtons & JOY_B_B1) != 0;
  ji->js[1].button[1] = (joyButtons & JOY_B_B2) != 0;

  if (tries == TIMER_READ_TRIES) {
    joyX1 = oJoyX1;
    joyY1 = oJoyY1;
    joyX2 = oJoyX2;
    joyY2 = oJoyY2;
  } // if

  ji->js[0].x = joyX1;
  ji->js[0].y = joyY1;
  ji->js[1].x = joyX2;
  ji->js[1].y = joyY2;

  if ((joyflags & JOY_A_X) && joyX1 >= joyMax) return 0;
  if ((joyflags & JOY_A_Y) && joyY1 >= joyMax) return 0;
  if ((joyflags & JOY_B_X) && joyX2 >= joyMax) return 0;
  if ((joyflags & JOY_B_Y) && joyY2 >= joyMax) return 0;
  if (i >= joyMax) return 0;

  return 1;

} // readJoystickTimer

// joystick.c
